"""
Lambda serialization for tikcle.

Provides the classes required for serialization of lambdas and functions.
"""

import new
import uuid
import types
import marshal
import __builtin__

from tickle import Ticklable


class LambdaTicklable(Ticklable):
    """
    Ticklable for lambda types.
    """
    @classmethod
    def save(cls, tickler, obj):
        """
        Overload of `Ticklable.save`.
        """
        if obj.__name__ != "<lambda>":
            return tickler.save_global(obj)
        return super(LambdaTicklable, cls).save(tickler, obj)

    @staticmethod
    def dis_cells(func):
        """
        Extract cell key/values from a lambda closure.

        func -- the function object.

        Returns a dictionary of key/value pairs.
        """
        values = {}
        for k, n in enumerate(func.__code__.co_freevars):
            cell = func.__closure__[k]
            obj = type(lambda: 0)(
                (lambda x: lambda: x)(0).__code__, {}, None, None, (cell,)
            )()
            values[n] = obj
        return values

    @staticmethod
    def make_cells(**values):
        """
        Create cell objects from key/value pairs.

        **values -- the values.

        Returns a tuple containing the cell objects.
        """
        uid = str(uuid.uuid4()).replace("-", "")
        name = "make_cells_%s" % (uid,)

        code = "def %s():\n" % (name,)
        for k in values:
            code += "  %s = %s[%s]\n" % (k, name, repr(k),)
        code += "  return lambda: (%s)" % (", ".join(values.keys()),)

        globals()[name] = values
        exec(code)
        cells = locals()[name]().__closure__
        del globals()[name]

        return cells

    def __getstate__(self):
        """
        Overload of `Ticklable.__getstate__`.
        """
        co_globals = {}
        for x in __builtin__.__dict__:
            if x not in not_allowed_globals:
                co_globals[x] = __builtin__.__dict__[x]

        for k, v in self.data.__globals__.items():
            if v is self.data:
                continue

            if not (isinstance(v, types.ModuleType) or
                    isinstance(v, types.FunctionType) or
                    isinstance(v, type) or
                    isinstance(v, object) or
                    type(v) in [int, long, float, str, unicode]):
                continue

            if k in ["__builtins__"]:
                continue

            co_globals[k] = v

        return (marshal.dumps(self.data.__code__),
                co_globals,
                self.data.__defaults__,
                self.dis_cells(self.data),)

    def __setstate__(self, data):
        """
        Overload of `Ticklable.__setstate__`.
        """
        co_code, co_globals, co_defaults, co_cells = data

        co_codeobj = marshal.loads(co_code)
        co_closures = self.make_cells(**co_cells)

        self.data = new.function(co_codeobj, co_globals, "<lambda>", co_defaults, co_closures)


# The list of allowed globals
not_allowed_globals = [
    'Ellipsis',
    'NotImplemented',
]



"""
Advanced serialization library for Python 2.7+.

This library allows one to serialize more advanced types, such as lambdas, modules
and builtins. It mirrors the `pickle` module with more advanced features.
"""

import types
import pickle
from StringIO import StringIO


# Project info
__version__ = '0.1.0'

# Extensions registration
dispatch = {}


class Ticklable(object):
    """
    Ticklable base class.
    """
    @classmethod
    def save(cls, tickler, obj):
        """
        Method called to create a picklable instance of the object.
        """
        return tickler.save_inst(cls(obj))

    def __getinitargs__(self):
        """
        Get the picklable init arguments.
        """
        return (None,)

    def __init__(self, obj):
        """
        Initialize the picklable.

        obj -- the object to pickle.
        """
        self.data = obj

    def __getstate__(self):
        """
        Get the serialized object data.
        """
        return None

    def __setstate__(self, data):
        """
        Inject the serialized data into the object.

        data -- the serialized data.
        """
        self.data = None


class Tickler(pickle.Pickler):
    """
    Custom pickler class with added debugging and logging.
    """
    def save(self, obj):
        """
        Overload of `pickle.Pickler.save`.
        """
        if type(obj).__module__ == "site":
            return pickle.Pickler.save(self, None)

        try:
            picklable = obj.__class__.__dict__["__picklable__"](obj)
            self.dispatch[obj.__class__] = picklable.save
        except (AttributeError, KeyError):
            pass

        result = pickle.Pickler.save(self, obj)

        return result


class Untickler(pickle.Unpickler):
    """
    Overload of `pickle.Unpickler` to support custom types.
    """
    def load_build(self):
        """
        Overload of `pickle.Unpickler.load_build`.
        """
        pickle.Unpickler.load_build(self)
        last = self.stack[-1]

        if isinstance(last, Ticklable):
            obj = self.stack.pop()
            self.stack.append(obj.data)


def dumps(obj):
    """
    Serialize an object.

    obj -- the object to serialize.

    Returns a serialized version of the object.
    """
    from tickle.builtins import TypeTicklable
    from tickle.builtins import ModuleTicklable
    from tickle.lambdas import LambdaTicklable

    local_dispatch = {
        types.LambdaType: LambdaTicklable,
        types.ModuleType: ModuleTicklable,
        types.EllipsisType: Ticklable,
        types.NotImplementedType: Ticklable,
        types.ClassType: TypeTicklable,
        types.TypeType: TypeTicklable,
    }

    stream = StringIO()
    tickler = Tickler(stream)
    for T, cls in dispatch.items() + local_dispatch.items():
        tickler.dispatch[T] = cls.save
    tickler.dump(obj)
    stream.seek(0)
    return stream.read()


def loads(data):
    """
    Unserializes an object.

    data -- the serialized object.

    Returns a copy of the original object.
    """
    stream = StringIO(data)
    untickler = Untickler(stream)
    untickler.dispatch[pickle.BUILD[0]] = Untickler.load_build
    return untickler.load()



#!/usr/bin/python2.7

from setuptools import setup
from setuptools import find_packages

import tickle as main


setup(
    name=main.__name__,
    version=main.__version__,

    description=main.__doc__.split('\n')[1].strip(),
    long_description=main.__doc__.strip(),
    url='https://bitbucket.org/overridelogic/tickle',
    author='OverrideLogic',
    author_email='info@overridelogic.com',
    maintainer='Francis Lacroix',
    maintainer_email='f@overridelogic.com',

    license='MIT',
    platforms=['any'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],

    packages=find_packages(exclude=['tests', 'tests.*']),
    provides=[main.__name__],

    python_requires='>=2.7',
    setup_requires=['flake8>=3.2.1'],
    tests_require=[],
    test_suite='tests',
)



"""
Builtins module for tickle.

This module provides support for serializing built-in types likes types and modules.
"""

import types

from tickle import Ticklable


class TypeTicklable(Ticklable):
    """
    Ticklable for built-in types.
    """
    @classmethod
    def save(cls, tickler, obj):
        """
        Method called to create a pickable instance of the object.
        """
        if obj.__module__ != "__builtin__":
            return tickler.save_global(obj)
        return super(TypeTicklable, cls).save(tickler, obj)

    def __getstate__(self):
        """
        Overload of `Ticklable.__getstate__`.
        """
        for k, v in types.__dict__.items():
            if v == self.data:
                return k

    def __setstate__(self, data):
        """
        Overload of `Ticklable.__setstate__`.
        """
        try:
            self.data = types.__dict__[data]
        except KeyError:
            pass


class ModuleTicklable(Ticklable):
    """
    Ticklable for module types.
    """
    def __getstate__(self):
        """
        Overload of `Ticklable.__getstate__`.
        """
        return self.data.__name__

    def __setstate__(self, data):
        """
        Overload of `Ticklable.__setstate__`.
        """
        parts = data.split(".")
        if len(parts) > 1:
            name = parts[0:-1]
            attr = parts[-1]
            obj = __import__(".".join(name), fromlist=[attr])
            self.data = getattr(obj, attr)
        else:
            self.data = __import__(parts[0])
